/* Yamaha YM2203 driver */

#include "emu.h"
#include "soundhw.h"
#include "fm.h"
#include "psg.h"

#define MAX_YM2203 4

int num_ym2203;
static AUDIOSTREAM *psg_stream[MAX_YM2203], *fm_stream[MAX_YM2203];

AUDIO_CHIP_DRIVER ym2203_interface = {
   "Yamaha YM2203",
   &ym2203_Init,
   &ym2203_Reset,
   &ym2203_Emulate,
   &ym2203_Shutdown
};

int ym2203_Init(int num, int clock, int rate, int buffer_size)
{
   int n;
   GUI_PrintLog(FALSE, "Initialising %i YM2203 chip(s)...", num);
  
   if (num > MAX_YM2203)
   {
      GUI_PrintLog(TRUE, "Current YM2203 driver cannot use more than %i chips.", MAX_YM2203);
      return FALSE;
   }
   num_ym2203 = num;

   GUI_PrintLog(FALSE, "Creating YM2203 audio stream(s).");
   for (n=0;n<num;n++)
   {
      fm_stream[n] = play_audio_stream(buffer_size, 8, FALSE, rate, 255, 128);
      psg_stream[n] = play_audio_stream(buffer_size, 8, FALSE, rate, 255, 128);
      if (!fm_stream[n] || !psg_stream[n])
      {
         GUI_PrintLog(TRUE, "Out of memory for FM emulator.");
         GUI_MessageBox(MB_OK, NULL, "Out of memory for FM emulator.");
         do {
            stop_audio_stream(fm_stream[n]);
            stop_audio_stream(psg_stream[n]);
         } while(--n);
         return FALSE;
      }
   }

   if (!OPNInit(num, clock, rate, buffer_size))
   {
      GUI_PrintLog(TRUE, "Out of memory for FM emulator.");
      GUI_MessageBox(MB_OK, NULL, "Out of memory for FM emulator.");
      for (n=0;n<num;n++)
      {
         stop_audio_stream(fm_stream[n]);
         stop_audio_stream(psg_stream[n]);
      }
      return FALSE;
   }

   for (n=0;n<num;n++)
   {
      if (!AYInit(n, rate, buffer_size))
      {
         GUI_PrintLog(TRUE, "Out of memory for PSG emulator.");
         GUI_MessageBox(MB_OK, NULL, "Out of memory for PSG emulator.");
         for (n=0;n<num;n++)
         {
            stop_audio_stream(fm_stream[n]);
            stop_audio_stream(psg_stream[n]);
         }
         return FALSE;
      }
      AYSetClock(n, clock);
   }

   GUI_PrintLog(FALSE, "YM2203 startup OK.");
   return TRUE;
}

void ym2203_Reset(int num)
{
   GUI_PrintLog(FALSE, "Resetting YM2203 chip %i.", num);
   AYResetChip(num);
   OPNResetChip(num);
}

void ym2203_Emulate(int num)
{
   byte *buffer;

   /* Update the PSG audio section */
   buffer = get_audio_stream_buffer(psg_stream[num]);
   if (buffer)
   {
      AYUpdate(num, buffer);
      free_audio_stream_buffer(psg_stream[num]);
   }

   /* Update the FM audio section */
   buffer = get_audio_stream_buffer(fm_stream[num]);
   if (buffer)
   {
      OPNSetBuffer(num, buffer);
      OPNUpdate();
      free_audio_stream_buffer(fm_stream[num]);
   }
}

void ym2203_Shutdown(void)
{
   int n;
   GUI_PrintLog(FALSE, "Shutting down YM2203 emulator.");

   for (n=0;n<num_ym2203;n++)
   {
      if (fm_stream[n]) stop_audio_stream(fm_stream[n]);
      if (psg_stream[n]) stop_audio_stream(psg_stream[n]);
      fm_stream[n] = psg_stream[n] = NULL;
   }

   OPNShutdown();
}

byte ym2203_ReadStatus(int chip)
{
   return OPNReadStatus(chip);
}

byte ym2203_ReadReg(int chip, int reg)
{
   if (reg < 16)
   {
      return AYReadReg(chip, reg);
   } else {
      return OPNReadStatus(chip);
   }
}

void ym2203_WriteReg(int chip, int reg, byte value)
{
   if (reg < 16)
   {
      AYWriteReg(chip, reg, value);
   } else {
      OPNWriteReg(chip, reg, value);
   }
}

void ym2203_SetPortHandler(int chip, int port, byte (*handler)(int,int,int,byte))
{
   AYSetPortHandler(chip, port+AY_PORTA, handler);
}
