/*
 * Performance monitoring library.
 * by D.Michel
 * 100% non portable :)
 *
 * NOTE: As far as I can tell, this uses PIT channel 2 to read the time from.
 *       So, Allegro reprogramming the timer shouldn't affect it, but using
 *       the PC speaker *will*, so don't use it.
 */

#include <stdio.h>
#include "clock.h"

Clock_t Clock[MAX_CLOCKS];

/*
 * starts a clock ticking.
 */
void Clock_Start(int n)
{
   Clock[n].start = Clock_GetTime();
   Clock[n].active = 1;
}

/*
 * stops a clock and sets clock->time to the time it ran for.
 * also sets: min = lowest duration so far
 *            max = highest duration so far
 *            total = total time altogether
 *            average = average duration each run
 */
void Clock_Stop(int n)
{
   int i;

   if (!Clock[n].active)
      return;

   Clock[n].frame++;
   Clock[n].stop = Clock_GetTime();
   Clock[n].active = 0;
   Clock[n].time = Clock_Diff(Clock[n].start, Clock[n].stop);
   Clock[n].time -= Clock[n].skip;
   Clock[n].skip = 0;
   Clock[n].min = MIN(Clock[n].min, Clock[n].time);
   Clock[n].max = MAX(Clock[n].max, Clock[n].time);
   Clock[n].total += Clock[n].time;
   Clock[n].average = Clock[n].total / Clock[n].frame;

   for (i=0;i<MAX_CLOCKS;i++)
   {
      if (Clock[i].active)
      {
         Clock[i].skip += Clock[n].time;
      }
   }
}

/*
 * calculate the time difference between a start and a stop.
 */
int Clock_Diff(int start, int stop)
{
   unsigned int t;

   if (start < stop)
      t = (start + 0x10000) - stop;
   else
      t = start - stop;

   return t;
}

/*
 * returns the current time.
 */
int Clock_GetTime(void)
{
   unsigned int t;

   DISABLE();
   outportb(0x43, 0x80);
   IO_DELAY();
   t = inportb(0x42);
   IO_DELAY();
   t |= inportb(0x42) << 8;
   IO_DELAY();
   ENABLE();

   return t;
}

/*
 * returns a time in a human-readable format.
 */
char *Clock_GetString(int time)
{
   static char string[32];
   int value, ms, us;

   value = (time * 838) / 1000;
   ms = (value / 1000);
   us = (value % 1000);

   sprintf(string, "%3i.%03i", ms, us);
   return string;
}

/*
 * starts up the clock system.
 */
void Clock_Init(void)
{
   DISABLE();
   outportb(0x43, 0xB4);
   IO_DELAY();
   outportb(0x42, 0xFF);
   IO_DELAY();
   outportb(0x42, 0xFF);
   IO_DELAY();
   outportb(0x61, (inportb(0x61) & 0xFC) | 0x01);
   ENABLE();
   Clock_Reset();
}

/*
 * resets all clocks back to their default state.
 */
void Clock_Reset(void)
{
   int i;

   for (i=0;i<MAX_CLOCKS;i++)
   {
      Clock[i].frame = 0;
      Clock[i].min = 999999999;
      Clock[i].max = 0;
      Clock[i].average = 0;
      Clock[i].skip = 0;
      Clock[i].total = 0;
      Clock[i].active = 0;
   }
}

/*
 * shuts down the clock system.
 */
void Clock_Exit(void)
{
   outportb(0x61, inportb(0x61) & 0xFC);
}
