#include <stdarg.h>
#include "gui.h"

WINDOW *top_window, *bottom_window, *active_window;
int (*startup_function)(void)=NULL, (*shutdown_function)(WINDOW *window, GADGET *d)=NULL;
PALETTE gui_palette;
int vsync_done, gui_palette_dirty;

int GUI_DispatchGlobalMessage(MSG *msg)
{
   WINDOW *window;
   switch(msg->type)
   {
   case MSG_START:
      GUI_PrintLog(FALSE, "Initalising desktop...");
      top_window = bottom_window = active_window = NULL;
      GUI_LoadConfig();

      GUI_Refresh(NULL);
      if (!GUI_InitLog()) return D_CLOSE;

      if (startup_function)
      {
         if (!startup_function()) return D_CLOSE;
      }

      GUI_PrintLog(FALSE, "Desktop startup OK.");
      return D_O_K;
   case MSG_IDLE:
      return GUI_ProcessWindowMessages();
   case MSG_REFRESH:
      GUI_Refresh(NULL);
      return D_O_K;
   case MSG_CLICK:
      window = GUI_GetWindowAtCoords(msg->x, msg->y);
      if (window)
      {
         GUI_SwitchFocus(window);

         msg->x -= window->x;
         msg->y -= window->y;
         GUI_PostMessage(&window->msg_queue, msg);
      }
      return D_O_K;
   case MSG_CHAR:
      if ((msg->key & 0xff) == 27)
      {
         GUI_PrintLog(FALSE, "Escape pressed, closing desktop.");
         return D_CLOSE;
      } else {
         if (active_window)
            GUI_PostMessage(&active_window->msg_queue, msg);
         return D_O_K;
      }
   case MSG_END:
      if (shutdown_function)
      {
         shutdown_function(NULL, NULL);
      }

      GUI_PrintLog(FALSE, "Closing all desktop windows...");
      GUI_SaveConfig();
      window = top_window;
      while(window)
      {
         GUI_PostMessage(&window->msg_queue, msg);
         window = window->next;
      }
      return GUI_ProcessWindowMessages();
   }
      
   return D_O_K;
}

void GUI_LoadConfig(void)
{
   int n,t;
   RGB rgb;

   GUI_PrintLog(FALSE, "Reading configuration from config file.");
   GUI_ReadConfigColour(GUI_BACKDROP,       "background",        0xa0dbff);
   GUI_ReadConfigColour(GUI_URGENT_TEXT,    "urgent_text",       0x0000ff);
   GUI_ReadConfigColour(GUI_DISABLED_TEXT,  "disabled_text",     0x808080);
   GUI_ReadConfigColour(GUI_BG_TEXT,        "normal_text",       0x404040);
   GUI_ReadConfigColour(GUI_BG_HIGHLIGHT,   "normal_highlight",  0xffffff);
   GUI_ReadConfigColour(GUI_BG_NORMAL,      "normal_window",     0xa0a0a0);
   GUI_ReadConfigColour(GUI_BG_SHADOW,      "normal_shadow",     0x606060);
   GUI_ReadConfigColour(GUI_FG_TEXT,        "active_title",      0x000000);
   GUI_ReadConfigColour(GUI_FG_HIGHLIGHT,   "active_highlight",  0xa0a0c0);
   GUI_ReadConfigColour(GUI_FG_NORMAL,      "active_window",     0x8080a0);
   GUI_ReadConfigColour(GUI_FG_SHADOW,      "active_shadow",     0x303040);

   rgb.r = rgb.g = rgb.b = 0;
   gui_palette[GUI_BLACK] = gui_palette[GUI_TRANSPARENT] = rgb;
   rgb.r = rgb.g = rgb.b = 63;
   gui_palette[GUI_WHITE] = rgb;

   for (n=GUI_BACKDROP_START;n<=GUI_BACKDROP_END;n++)
   {
      t = 16+ 47*(n - GUI_BACKDROP_START) / (GUI_BACKDROP_END + 1 - GUI_BACKDROP_START);
      rgb.r = (t * gui_palette[GUI_BACKDROP].r) / 64;
      rgb.g = (t * gui_palette[GUI_BACKDROP].g) / 64;
      rgb.b = (t * gui_palette[GUI_BACKDROP].b) / 64;
      gui_palette[n] = rgb;
   }

   set_palette(gui_palette);
}

void GUI_ReadConfigColour(int colour, char *name, int def)
{
   int c;
   RGB rgb;

   c = get_config_hex("GUI_colours", name, def);

   rgb.r = ((c >> 16) & 255) / 4;
   rgb.g = ((c >>  8) & 255) / 4;
   rgb.b = ( c        & 255) / 4;
   gui_palette[colour] = rgb;
}

void GUI_SaveConfig(void)
{
   GUI_PrintLog(FALSE, "Saving configuration back to config file.");
   GUI_SaveConfigColour(GUI_BACKDROP,       "background");
   GUI_SaveConfigColour(GUI_URGENT_TEXT,    "urgent_text");
   GUI_SaveConfigColour(GUI_DISABLED_TEXT,  "disabled_text");
   GUI_SaveConfigColour(GUI_BG_TEXT,        "normal_text");
   GUI_SaveConfigColour(GUI_BG_HIGHLIGHT,   "normal_highlight");
   GUI_SaveConfigColour(GUI_BG_NORMAL,      "normal_window");
   GUI_SaveConfigColour(GUI_BG_SHADOW,      "normal_shadow");
   GUI_SaveConfigColour(GUI_FG_TEXT,        "active_title");
   GUI_SaveConfigColour(GUI_FG_HIGHLIGHT,   "active_highlight");
   GUI_SaveConfigColour(GUI_FG_NORMAL,      "active_window");
   GUI_SaveConfigColour(GUI_FG_SHADOW,      "active_shadow");
}

void GUI_SaveConfigColour(int colour, char *name)
{
   int c;
   RGB rgb;

   rgb = gui_palette[colour];

   c = ((rgb.r * 4) << 16)
     | ((rgb.g * 4) <<  8)
     | ( rgb.b * 4);

   set_config_hex("GUI_colours", name, c);
}

void GUI_BringWindowToFront(WINDOW *window)
{
   if (window == top_window) return;
   if (window == bottom_window) bottom_window = window->prev;

   if (window->next) window->next->prev = window->prev;
   if (window->prev) window->prev->next = window->next;

   window->next = top_window;
   window->prev = NULL;
   top_window->prev = window;

   top_window = window;
}

void GUI_SetStartupFunction(int (*startup)(void), int (*shutdown)(WINDOW *window, GADGET *d))
{
   startup_function = startup;
   shutdown_function = shutdown;
}

int GUI_ProcessMessage(WINDOW *window, MSG *msg)
{
   GADGET *gadget;
   int result;

   switch(msg->type)
   {
   case MSG_CLICK:
      /* get gadget under cursor and make co-ords relative to it */
      gadget = GUI_GetGadgetAtCoords(window, msg->x, msg->y);

      GUI_SwitchInputFocus(window, gadget);

      msg->x -= gadget->x;
      msg->y -= gadget->y;
      return gadget->proc(msg, window, gadget);
   case MSG_CHAR:
      gadget = window->input_focus;
      if (gadget)
      {
         return gadget->proc(msg, window, gadget);
      }
      return D_O_K;
   default:
      gadget = window->gadgets;
      result = D_O_K;
      while(gadget->proc)
      {
         result |= gadget->proc(msg, window, gadget);
         gadget++;
      }

      if (msg->type == MSG_REFRESH) GUI_Refresh(window);
      return result;
   }
}

int GUI_ProcessWindowMessages(void)
{
   WINDOW *window, *prev;
   MSG msg;
   int result, lastmsg;

   if (active_window == NULL)
   {
      GUI_SwitchFocus(top_window);
   }

   /* start from window with lowest z-order */
   window = bottom_window;
   while(window)
   {
      prev = window->prev;

      /* post an IDLE message */
      msg.type = MSG_IDLE;
      GUI_PostMessage(&window->msg_queue, &msg);

      GUI_CheckWindowForRedraw(window);

      lastmsg = 0;
      while (GUI_GetMessage(&window->msg_queue, &msg))
      {
         /* don't refresh more than once */
         if (lastmsg == MSG_REFRESH && msg.type == MSG_REFRESH) continue;

         lastmsg = msg.type;
         /* process waiting messages */
         if (msg.type == MSG_END)
         {
            result = GUI_ProcessMessage(window, &msg);

            /* remove window and ignore the rest of it's messages */
            GUI_RemoveWindow(window);
            break;
         } else {
            result = GUI_ProcessMessage(window, &msg);
            if (result & D_CLOSE)
            {
               /* send quit message immediately */
               msg.type = MSG_END;
               GUI_ProcessMessage(window, &msg);

               /* remove window and ignore the rest of it's messages */
               GUI_RemoveWindow(window);
               break;   
            }
         }
      }

      window = prev;
   }

   return D_O_K;
}

void GUI_Refresh(WINDOW *finalwindow)
{
   WINDOW *window;
   RECT cliprect;

   window = top_window;

   vsync_done = FALSE;
   cliprect.x = cliprect.y = 0;
   cliprect.x2 = SCREEN_W - 1;
   cliprect.y2 = SCREEN_H - 1;
   GUI_RecursiveRefresh(window, cliprect, finalwindow);

   if (gui_palette_dirty)
   {
      //vsync();
      set_palette(gui_palette);
      gui_palette_dirty = FALSE;
   }
}

void GUI_RecursiveRefresh(WINDOW *window, RECT cliprect, WINDOW *finalwindow)
{
   RECT dest, newclip;
   WINDOW *next;
   int x,y,mx, my, mx2, my2, mouse_in_window;

   /* first draw this window (inside the given clipping window) */
   if (window)
   {
      dest.x = window->x;
      dest.y = window->y;
      dest.x2 = dest.x + window->w - 1;
      dest.y2 = dest.y + window->h - 1;
   } else {
      /* draw desktop background */
      dest.x = 0;
      dest.y = 0;
      dest.x2 = SCREEN_W - 1;
      dest.y2 = SCREEN_H - 1;
   }
   if (dest.x < cliprect.x) dest.x = cliprect.x;
   if (dest.y < cliprect.y) dest.y = cliprect.y;
   if (dest.x2 > cliprect.x2) dest.x2 = cliprect.x2;
   if (dest.y2 > cliprect.y2) dest.y2 = cliprect.y2;

   /* don't draw this window if no drawing area left */
   if (dest.x <= dest.x2 && dest.y <= dest.y2)
   {
      if (window == finalwindow || finalwindow == NULL)
      {
         mx = mouse_x - gui_mouse_focus_x;
         my = mouse_y - gui_mouse_focus_y;
         mx2 = mx + gui_mouse_bitmap->w - 1;
         my2 = my + gui_mouse_bitmap->h - 1;

         if (mx2 < dest.x || my2 < dest.y || mx > dest.x2 || my >= dest.y2)
         {
            mouse_in_window = FALSE;
         } else {
            mouse_in_window = TRUE;
            scare_mouse();
         }

         if (window)
         {
            if (!vsync_done && window->flags & W_VSYNC)
            {
               if (gui_palette_dirty)
               {
                  set_palette(gui_palette);
                  gui_palette_dirty = FALSE;
               } else {
                  vsync();
               }
               vsync_done = TRUE;
            }

            blit(window->bmp, screen,
                 dest.x - window->x,
                 dest.y - window->y,
                 dest.x,
                 dest.y,
                 dest.x2 - dest.x + 1,
                 dest.y2 - dest.y + 1);

            //rect(screen, dest.x, dest.y, dest.x2, dest.y2,
            //     makecol(rand()&255,rand()&255,rand()&255));
         } else {
            x = dest.x - (SCREEN_W - gui_backdrop->w)/2;
            y = dest.y - (SCREEN_H - gui_backdrop->h)/2;
            blit(gui_backdrop, screen, x,y, dest.x, dest.y, dest.x2+1-dest.x, dest.y2+1-dest.y);
            //rectfill(screen, dest.x, dest.y, dest.x2, dest.y2, GUI_BACKDROP);
            //rect(screen, dest.x, dest.y, dest.x2, dest.y2,
            //     makecol(rand()&255,rand()&255,rand()&255));
         }
   
         if (mouse_in_window) unscare_mouse();
      }
   }

   /* now recurse into all the lower windows */

   /* don't recurse down if this is the last window to do */
   if (window == finalwindow) return;

   next = window->next;

   /* everything to the left (and top-left / bottom-left) */
   newclip = cliprect;
   newclip.x2 = MIN(window->x - 1, cliprect.x2);
   if (newclip.x2 >= newclip.x)
      GUI_RecursiveRefresh(next, newclip, finalwindow);

   /* chop the bit we just did off the clipping area
      (stop if no clipping area remains) */
   cliprect.x = MAX(window->x, cliprect.x);
   if (cliprect.x > cliprect.x2) return;

   /* everything to the right (and top-right / bottom-right)*/
   newclip = cliprect;
   newclip.x = MAX(window->x + window->w, cliprect.x);
   if (newclip.x <= newclip.x2)
      GUI_RecursiveRefresh(next, newclip, finalwindow);

   /* chop the bit we just did off the clipping area
      (stop if no clipping area remains) */
   cliprect.x2 = MIN(window->x + window->w - 1, cliprect.x2);
   if (cliprect.x2 < cliprect.x) return;

   /* everything to the top */
   newclip = cliprect;
   newclip.y2 = MIN(window->y - 1, cliprect.y2);
   if (newclip.y2 >= newclip.y)
      GUI_RecursiveRefresh(next, newclip, finalwindow);

   /* chop the bit we just did off the clipping area
      (stop if no clipping area remains) */
   cliprect.y = MAX(window->y, cliprect.y);
   if (cliprect.y > cliprect.y2) return;

   /* everything to the bottom */
   newclip = cliprect;
   newclip.y = MAX(window->y + window->h, cliprect.y);
   if (newclip.y <= newclip.y2)
      GUI_RecursiveRefresh(next, newclip, finalwindow);
}

void GUI_CheckWindowForRedraw(WINDOW *window)
{
   GADGET *gadget;
   MSG msg;
   int refresh_needed;

   msg.type = MSG_DRAW;
   refresh_needed = FALSE;

   gadget = window->gadgets;
   while(gadget->proc)
   {
      if (gadget->flags & D_DIRTY)
      {
         gadget->flags &= ~D_DIRTY;
         GUI_SendGadgetMessage(window, gadget, &msg);
         refresh_needed = TRUE;
      }
      gadget++;
   }

   if (refresh_needed)
   {
      msg.type = MSG_REFRESH;
      GUI_PostMessage(&window->msg_queue, &msg);
   }
}

void GUI_SendGadgetMessage(WINDOW *window, GADGET *gadget, MSG *msg)
{
   gadget->proc(msg, window, gadget);
}

WINDOW *GUI_GetWindowAtCoords(int x, int y)
{
   WINDOW *window;
   int wx, wy, wx2, wy2;
   window = top_window;

   while(window)
   {
      wx = window->x;
      wy = window->y;
      wx2 = wx + window->w;
      wy2 = wy + window->h;

      if (x >= wx && y >= wy && x < wx2 && y < wy2) return window;
      window = window->next;
   }

   return NULL;
}

GADGET *GUI_GetGadgetAtCoords(WINDOW *window, int x, int y)
{
   GADGET *gadget;
   int gx,gy, gx2,gy2;

   /* skip to the last gadget */
   gadget = window->gadgets;
   while(gadget->proc) gadget++;
   gadget--;

   /* work backwards till we find the top-most object under the cursor */
   while(1)
   {
      gx = gadget->x;
      gy = gadget->y;
      gx2 = gx + gadget->w;
      gy2 = gy + gadget->h;

      if (x >= gx && y >= gy && x < gx2 && y < gy2) return gadget;
      gadget--;
   }                    
}

void GUI_RemoveWindow(WINDOW *window)
{
   GUI_PrintLog(FALSE, "Removing window.");
   if (window->close_func) window->close_func(window);

   if (window == active_window) active_window = NULL;

   if (window->prev) window->prev->next = window->next;
   if (window->next) window->next->prev = window->prev;
   if (window == top_window) top_window = window->next;
   if (window == bottom_window) bottom_window = window->prev;

   destroy_bitmap(window->bmp);
   while (window->gadgets->proc)
   {
      destroy_bitmap(window->gadgets->bmp);
      window->gadgets++;
   }
   free(window);
   GUI_Refresh(NULL);
}

void GUI_MoveWindow(WINDOW *window, int x, int y)
{
   window->x = x;
   window->y = y;
   GUI_Refresh(NULL);
}

WINDOW *GUI_NewWindow(int x, int y, GADGET *gadgets, void (*closefunc)(WINDOW *window))
{
   WINDOW *window;
   GADGET *gadget;
   MSG msg;

   /* set up window */
   window = (WINDOW *)malloc(sizeof(WINDOW));
   if (!window)
   {
      GUI_PrintLog(TRUE, "Out of memory for new window.");
      GUI_MessageBox(MB_OK, NULL, "Out of memory for new window.");
      return NULL;
   }

   window->x = x;
   window->y = y;
   window->w = gadgets->w;
   window->h = gadgets->h;
   window->gadgets = gadgets;
   window->flags = 0;
   window->input_focus = NULL;
   window->close_func = closefunc;

   window->bmp = create_bitmap(window->w, window->h);
   if (!window->bmp)
   {
      GUI_PrintLog(TRUE, "Out of memory for new window bitmap.");
      GUI_MessageBox(MB_OK, NULL, "Out of memory for new window bitmap.");
      free(window);
      return NULL;
   }

   window->msg_queue.head = window->msg_queue.tail = 0;

   /* set up all the gadgets */
   gadget = window->gadgets;
   while(gadget->proc)
   {
      gadget->bmp = create_sub_bitmap(window->bmp,
                                      gadget->x, gadget->y,
                                      gadget->w, gadget->h);
      if (!gadget->bmp)
      {
         GUI_PrintLog(TRUE, "Out of memory for window gadgets.");
         GUI_MessageBox(MB_OK, NULL, "Out of memory for window gadgets.");
         /* free all the gadgets */
         while(gadget >= window->gadgets)
         {
            destroy_bitmap(gadget->bmp);
            gadget--;
         }
         destroy_bitmap(window->bmp);
         free(window);
         return FALSE;
      }

      gadget++;
   }

   /* add window into the window list (highest z-order) */
   if (top_window)
   {
      window->next = top_window;
      window->next->prev = window;

      window->prev = NULL;

      top_window = window;
   } else {
      top_window = bottom_window = window;
      window->next = window->prev = NULL;
   }

   /* post the initial messages to the window */
   msg.type = MSG_START;
   GUI_PostMessage(&window->msg_queue, &msg);
   msg.type = MSG_DRAW;
   GUI_PostMessage(&window->msg_queue, &msg);
   msg.type = MSG_REFRESH;
   GUI_PostMessage(&window->msg_queue, &msg);

   /* make this window the active one */
   GUI_SwitchFocus(window);

   GUI_PrintLog(FALSE, "Window made successfully.");
   return window;
}

int GUI_MessageBox(int flags, char *msg2, char *msg, ...)
{
   RECT win, b1, b2;
   MSG newmsg;
   int mb_w, w,h;
   char text[300];
   va_list ap;
  
   va_start(ap, msg);
   vsprintf(text, msg, ap);
   va_end(ap);

   mb_w = 500;
   win.x = (SCREEN_W - mb_w) / 2;
   win.y = (SCREEN_H - 100) / 2;
   win.x2 = win.x + mb_w;
   win.y2 = win.y + 100;

   GUI_SetMouseCursor(gui_data[GUI_CURSOR].dat, 1, 1);
   scare_mouse();
   GUI_DrawRect3D(TRUE, screen, win.x, win.y, win.x2, win.y2);
   rectfill(screen, win.x+1, win.y+1, win.x2-1, win.y2-1, GUI_FG_NORMAL);
   GUI_DrawRect3D(TRUE, screen, win.x2-3, win.y2-3, win.x+3, win.y+3);
   rectfill(screen, win.x+4, win.y+4, win.x2-4, win.y2-4, GUI_BG_NORMAL);

   text_mode(GUI_BG_NORMAL);
   if (flags == MB_OK)
   {
      w = 30;
      h = 30;
      b1.x = win.x + (mb_w - w) / 2;
      b1.y = win.y + (100 - h) - 10;
      b1.x2 = b1.x + w - 1;
      b1.y2 = b1.y + h - 1;
      GUI_DrawRect3D(FALSE, screen, b1.x+1, b1.y+1, b1.x2-1, b1.y2-1);
      rect(screen, b1.x, b1.y, b1.x2, b1.y2, GUI_BLACK);

      gui_textout(screen, "OK", (b1.x+b1.x2)/2, b1.y+9, GUI_BLACK, TRUE);
   }
   if (flags == MB_YESNO)
   {
      w = 30;
      h = 30;
      b1.x = win.x + (mb_w - w) / 3;
      b1.y = win.y + (100 - h) - 10;
      b1.x2 = b1.x + w - 1;
      b1.y2 = b1.y + h - 1;

      b2.x = win.x + ((mb_w - w) * 2/3);
      b2.y = b1.y;
      b2.x2 = b2.x + w - 1;
      b2.y2 = b2.y + h - 1;
      GUI_DrawRect3D(FALSE, screen, b1.x+1, b1.y+1, b1.x2-1, b1.y2-1);
      rect(screen, b1.x, b1.y, b1.x2, b1.y2, GUI_BLACK);
      GUI_DrawRect3D(FALSE, screen, b2.x, b2.y, b2.x2, b2.y2);

      gui_textout(screen, "Yes", (b1.x+b1.x2)/2, b1.y+9, GUI_BLACK, TRUE);
      gui_textout(screen, "No", (b2.x+b2.x2)/2, b2.y+9, GUI_BLACK, TRUE);
   }

   gui_textout(screen, text, (win.x+win.x2)/2, win.y+14, GUI_BLACK, TRUE);
   if (msg2)
      gui_textout(screen, msg2, (win.x+win.x2)/2, win.y+30, GUI_BLACK, TRUE);

   unscare_mouse();

   /* post a global refresh message for afterwards */
   newmsg.type = MSG_REFRESH;
   GUI_PostMessage(&global_msg_queue, &newmsg);

   while(1)
   {
      if (keypressed())
      {
         if (flags == MB_OK)
         {
            readkey();
            return 1;
         }
         if (flags == MB_YESNO)
         {
            switch(readkey()&0xff)
            {
               case 'y': case 'Y': return 1;
               case 'n': case 'N': return 2;
               case 13: case 10: case ' ': return 1;
               case 27: return 2;
            }
         }
      }
      if (mouse_b)
      {
         if (mouse_x >= b1.x && mouse_y >= b1.y
          && mouse_x <= b1.x2 && mouse_y <= b1.y2)
         {
            scare_mouse();
            GUI_DrawRect3D(FALSE, screen, b1.x2-1, b1.y2-1, b1.x+1, b1.y+1);
            unscare_mouse();
            while(mouse_b);
            return 1;
         }

         if (flags == MB_YESNO)
         {
            if (mouse_x >= b2.x && mouse_y >= b2.y
             && mouse_x <= b2.x2 && mouse_y <= b2.y2)
            {
               /* display button 2 as default */
               scare_mouse();
               rect(screen, b1.x, b1.y, b1.x2, b1.y2, GUI_BG_NORMAL);
               rect(screen, b2.x, b2.y, b2.x2, b2.y2, GUI_BLACK);
               GUI_DrawRect3D(FALSE, screen, b2.x2-1, b2.y2-1, b2.x+1, b2.y+1);
               unscare_mouse();
               while(mouse_b);
               return 2;
            }
         }
      }
   }
}

void GUI_SwitchInputFocus(WINDOW *window, GADGET *gadget)
{
   MSG msg;
   if (gadget != window->input_focus)
   {
      msg.type = MSG_WANTFOCUS;
      if (gadget->proc(&msg, window, gadget) == D_WANTFOCUS)
      {
         msg.type = MSG_LOSTFOCUS;
         if (window->input_focus)
         {
            window->input_focus->flags |= D_DIRTY;
            window->input_focus->proc(&msg, window, window->input_focus);
         }

         window->input_focus = gadget;

         msg.type = MSG_GOTFOCUS;
         gadget->flags |= D_DIRTY;
         gadget->proc(&msg, window, gadget);
      }
   }
}

void GUI_SwitchFocus(WINDOW *newwindow)
{
   MSG newmsg;
   if (active_window == newwindow) return;

   if (active_window)
   {
      newmsg.type = MSG_DRAW;
      GUI_PostMessage(&active_window->msg_queue, &newmsg);
      newmsg.type = MSG_REFRESH;
      GUI_PostMessage(&active_window->msg_queue, &newmsg);
   }

   active_window = newwindow;

   GUI_BringWindowToFront(newwindow);

   newmsg.type = MSG_DRAW;
   GUI_PostMessage(&newwindow->msg_queue, &newmsg);
   newmsg.type = MSG_REFRESH;
   GUI_PostMessage(&newwindow->msg_queue, &newmsg);
}

void GUI_DrawRect3D(int fg, BITMAP *bmp, int x1, int y1, int x2, int y2)
{
   if (fg)
   {
      hline(bmp, x1, y2, x2, GUI_FG_SHADOW);
      vline(bmp, x2, y1, y2, GUI_FG_SHADOW);
      hline(bmp, x1, y1, x2, GUI_FG_HIGHLIGHT);
      vline(bmp, x1, y1, y2, GUI_FG_HIGHLIGHT);
   } else {
      hline(bmp, x1, y2, x2, GUI_BG_SHADOW);
      vline(bmp, x2, y1, y2, GUI_BG_SHADOW);
      hline(bmp, x1, y1, x2, GUI_BG_HIGHLIGHT);
      vline(bmp, x1, y1, y2, GUI_BG_HIGHLIGHT);
   }
}

void GUI_DrawLine3D(int fg, BITMAP *bmp, int x1, int y1, int x2, int y2)
{
   if (fg)
   {
      line(bmp, x1, y1, x2, y2, GUI_FG_SHADOW);
      line(bmp, x1, y1+1, x2, y2+1, GUI_FG_HIGHLIGHT);
   } else {
      line(bmp, x1, y1, x2, y2, GUI_BG_SHADOW);
      line(bmp, x1, y1+1, x2, y2+1, GUI_BG_HIGHLIGHT);
   }
}

void GUI_DrawDottedRect(BITMAP *bmp, int x1, int y1, int x2, int y2)
{
   int x,y;
   for (x=x1;x<x2;x+=2)
   {
      _putpixel(bmp, x, y1, GUI_BLACK);
      _putpixel(bmp, x+1, y2, GUI_BLACK);
   }
   for (y=y1;y<y2;y+=2)
   {
      _putpixel(bmp, x1, y, GUI_BLACK);
      _putpixel(bmp, x2, y+1, GUI_BLACK);
   }
}
